/*
 * Copyright (C) BABEC. All rights reserved.
 * Copyright (C) THL A29 Limited, a Tencent company. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

package blockcontract

import (
	"encoding/hex"
	"encoding/json"
	"errors"
	"fmt"
	"math"
	"reflect"
	"strconv"
	"strings"

	"chainmaker.org/chainmaker/utils/v2"

	"chainmaker.org/chainmaker/vm-native/v2/common"

	"chainmaker.org/chainmaker/common/v2/crypto/hash"
	"chainmaker.org/chainmaker/localconf/v2"
	commonPb "chainmaker.org/chainmaker/pb-go/v2/common"
	discoveryPb "chainmaker.org/chainmaker/pb-go/v2/discovery"
	storage "chainmaker.org/chainmaker/pb-go/v2/store"
	"chainmaker.org/chainmaker/pb-go/v2/syscontract"
	"chainmaker.org/chainmaker/protocol/v2"
	"github.com/gogo/protobuf/proto"
)

const (
	paramNameBlockHeight = "blockHeight"
	paramNameWithRWSet   = "withRWSet"
	paramNameBlockHash   = "blockHash"
	paramNameTxId        = "txId"
	// TRUE string const
	TRUE = "true"
)

var (
	logTemplateMarshalBlockInfoFailed = "marshal block info failed, %s"
	errStoreIsNil                     = fmt.Errorf("store is nil")
	//errDataNotFound                   = fmt.Errorf("data not found")
)

type BlockHeight int64
type BlockContract struct {
	methods map[string]common.ContractFunc
	log     protocol.Logger
}

func NewBlockContract(log protocol.Logger) *BlockContract {
	return &BlockContract{
		log:     log,
		methods: registerBlockContractMethods(log),
	}
}

func (c *BlockContract) GetMethod(methodName string) common.ContractFunc {
	return c.methods[methodName]
}

func registerBlockContractMethods(log protocol.Logger) map[string]common.ContractFunc {
	q := make(map[string]common.ContractFunc, 64)
	b := &BlockRuntime{log: log}

	q[syscontract.ChainQueryFunction_GET_BLOCK_BY_HEIGHT.String()] = common.WrapResultFunc(b.GetBlockByHeight)
	q[syscontract.ChainQueryFunction_GET_BLOCK_WITH_TXRWSETS_BY_HEIGHT.String()] = common.WrapResultFunc(
		b.GetBlockWithTxRWSetsByHeight)
	q[syscontract.ChainQueryFunction_GET_BLOCK_BY_HASH.String()] = common.WrapResultFunc(b.GetBlockByHash)
	q[syscontract.ChainQueryFunction_GET_BLOCK_WITH_TXRWSETS_BY_HASH.String()] = common.WrapResultFunc(
		b.GetBlockWithTxRWSetsByHash)
	q[syscontract.ChainQueryFunction_GET_BLOCK_BY_TX_ID.String()] = common.WrapResultFunc(b.GetBlockByTxId)
	q[syscontract.ChainQueryFunction_GET_TX_BY_TX_ID.String()] = common.WrapResultFunc(b.GetTxByTxId)
	q[syscontract.ChainQueryFunction_GET_LAST_CONFIG_BLOCK.String()] = common.WrapResultFunc(b.GetLastConfigBlock)
	q[syscontract.ChainQueryFunction_GET_LAST_BLOCK.String()] = common.WrapResultFunc(b.GetLastBlock)
	q[syscontract.ChainQueryFunction_GET_CHAIN_INFO.String()] = common.WrapResultFunc(b.GetChainInfo)
	q[syscontract.ChainQueryFunction_GET_NODE_CHAIN_LIST.String()] = common.WrapResultFunc(b.GetNodeChainList)
	q[syscontract.ChainQueryFunction_GET_FULL_BLOCK_BY_HEIGHT.String()] = common.WrapResultFunc(b.GetFullBlockByHeight)
	q[syscontract.ChainQueryFunction_GET_BLOCK_HEIGHT_BY_TX_ID.String()] = common.WrapResultFunc(b.GetBlockHeightByTxId)
	q[syscontract.ChainQueryFunction_GET_BLOCK_HEIGHT_BY_HASH.String()] = common.WrapResultFunc(b.GetBlockHeightByHash)
	q[syscontract.ChainQueryFunction_GET_BLOCK_HEADER_BY_HEIGHT.String()] = common.WrapResultFunc(
		b.GetBlockHeaderByHeight)
	q[syscontract.ChainQueryFunction_GET_ARCHIVED_BLOCK_HEIGHT.String()] = common.WrapResultFunc(
		b.GetArchiveBlockHeight)
	q[syscontract.ChainQueryFunction_GET_MERKLE_PATH_BY_TX_ID.String()] = common.WrapResultFunc(b.GetMerklePathByTxId)
	return q
}

type BlockRuntime struct {
	log protocol.Logger
}

type BlockRuntimeParam struct {
	height    uint64
	withRWSet string
	hash      string
	txId      string
}

// GetNodeChainList return list of chain
func (r *BlockRuntime) GetNodeChainList(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error
	var chainListBytes []byte

	// check params
	if _, err = r.validateParams(parameters); err != nil {
		return nil, err
	}

	blockChainConfigs := localconf.ChainMakerConfig.GetBlockChains()
	chainIds := make([]string, len(blockChainConfigs))
	for i, blockChainConfig := range blockChainConfigs {
		chainIds[i] = blockChainConfig.ChainId
	}

	chainList := &discoveryPb.ChainList{
		ChainIdList: chainIds,
	}
	chainListBytes, err = proto.Marshal(chainList)
	if err != nil {
		errMsg = fmt.Sprintf("marshal chain list failed, %s", err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return chainListBytes, nil
}

func (r *BlockRuntime) GetChainInfo(txSimContext protocol.TxSimContext, parameters map[string][]byte) ([]byte, error) {
	var errMsg string
	var err error

	// check params
	if _, err = r.validateParams(parameters); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var nodes []*discoveryPb.Node

	if block, err = r.getBlockByHeight(store, chainId, math.MaxUint64); err != nil {
		return nil, err
	}

	var provider protocol.ChainNodesInfoProvider
	if provider, err = txSimContext.GetChainNodesInfoProvider(); err != nil {
		r.log.Warn(err)
	} else {
		if nodes, err = r.getChainNodeInfo(provider, chainId); err != nil {
			return nil, err
		}
	}

	chainInfo := &discoveryPb.ChainInfo{
		BlockHeight: block.Header.BlockHeight,
		NodeList:    nodes,
	}

	chainInfoBytes, err := proto.Marshal(chainInfo)
	if err != nil {
		errMsg = fmt.Sprintf("marshal chain info failed, %s", err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return chainInfoBytes, nil
}

func (r *BlockRuntime) GetBlockByHeight(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameBlockHeight, paramNameWithRWSet); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByHeight(store, chainId, param.height); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if strings.ToLower(param.withRWSet) == TRUE {
		if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
			return nil, err
		}
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err := proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetBlockWithTxRWSetsByHeight(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameBlockHeight); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByHeight(store, chainId, param.height); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
		return nil, err
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err := proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetBlockByHash(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameBlockHash, paramNameWithRWSet); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByHash(store, chainId, param.hash); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if strings.ToLower(param.withRWSet) == TRUE {
		if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
			return nil, err
		}
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err := proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetBlockWithTxRWSetsByHash(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameBlockHash); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByHash(store, chainId, param.hash); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
		return nil, err
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err := proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetBlockByTxId(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameTxId, paramNameWithRWSet); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByTxId(store, chainId, param.txId); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if strings.ToLower(param.withRWSet) == TRUE {
		if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
			return nil, err
		}
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err := proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetMerklePathByTxId(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error
	var merkleTree [][]byte
	var merklePathsBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameTxId); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	chainConfig, err := store.GetLastChainConfig()
	if err != nil {
		return nil, err
	}
	hashType := chainConfig.Crypto.Hash

	var block *commonPb.Block

	if block, err = r.getBlockByTxId(store, chainId, param.txId); err != nil {
		return nil, err
	}

	hashes := make([][]byte, len(block.Txs))
	for i, tx := range block.Txs {
		var txHash []byte
		txHash, err = utils.CalcTxHash(hashType, tx)
		if err != nil {
			return nil, err
		}
		hashes[i] = txHash
	}

	merkleTree, err = hash.BuildMerkleTree(hashType, hashes)
	if err != nil {
		return nil, err
	}

	merklePaths := make([][]byte, 0)
	hash.GetMerklePath(hashType, []byte(param.txId), merkleTree, &merklePaths, false)

	merklePathsBytes, err = json.Marshal(merklePaths)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}

	return merklePathsBytes, nil

}

func (r *BlockRuntime) GetLastConfigBlock(txSimContext protocol.TxSimContext, parameters map[string][]byte) (
	[]byte, error) {
	var errMsg string
	var err error
	var blockInfoBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameWithRWSet); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getLastConfigBlock(store, chainId); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if strings.ToLower(param.withRWSet) == TRUE {
		if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
			return nil, err
		}
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err = proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetLastBlock(txSimContext protocol.TxSimContext, parameters map[string][]byte) ([]byte, error) {
	var errMsg string
	var err error
	var blockInfoBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(parameters, paramNameWithRWSet); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var block *commonPb.Block
	var txRWSets []*commonPb.TxRWSet

	if block, err = r.getBlockByHeight(store, chainId, math.MaxUint64); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(block, txSimContext); err != nil {
		return nil, err
	}

	if strings.ToLower(param.withRWSet) == TRUE {
		if txRWSets, err = r.getTxRWSetsByBlock(store, chainId, block); err != nil {
			return nil, err
		}
	}

	blockInfo := &commonPb.BlockInfo{
		Block:     block,
		RwsetList: txRWSets,
	}
	blockInfoBytes, err = proto.Marshal(blockInfo)
	if err != nil {
		errMsg = fmt.Sprintf(logTemplateMarshalBlockInfoFailed, err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	return blockInfoBytes, nil

}

func (r *BlockRuntime) GetTxByTxId(txSimContext protocol.TxSimContext, parameters map[string][]byte) ([]byte, error) {
	//var errMsg string
	var err error
	//var transactionInfoBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParamsAndLen(parameters, false, paramNameTxId); err != nil {
		return nil, err
	}
	if withRWSet, ok := parameters[paramNameWithRWSet]; ok {
		param.withRWSet = string(withRWSet)
	}
	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	if strings.ToLower(param.withRWSet) == "true" {
		txInfoWithRWSet, err1 := r.getTxInfoWithRWSetByTxId(store, chainId, param.txId)
		if err1 != nil {
			return nil, err1
		}
		return txInfoWithRWSet.Marshal()
	}
	txInfo, err := r.getTxWithInfoByTxId(store, chainId, param.txId)
	if err != nil {
		return nil, err
	}
	return txInfo.Marshal()

	//var tx *commonPb.Transaction
	//var block *commonPb.Block
	//
	//if tx, err = r.getTxByTxId(store, chainId, param.txId); err != nil {
	//	return nil, err
	//}
	//
	//if block, err = r.getBlockByTxId(store, chainId, param.txId); err != nil {
	//	return nil, err
	//}
	//
	//transactionInfo := &commonPb.TransactionInfo{
	//	Transaction: tx,
	//	BlockHeight: uint64(block.Header.BlockHeight),
	//}
	//
	//TODO:Devin 什么情况下要检查权限，从而决定只返回区块高度还是包含其他信息呢？
	//if transactionInfo, err = checkRoleAndGenerateTransactionInfo(txSimContext, transactionInfo); err != nil {
	//	return nil, err
	//}
	//
	//transactionInfoBytes, err = proto.Marshal(transactionInfo)
	//if err != nil {
	//	errMsg = fmt.Sprintf("marshal tx failed, %s", err.Error())
	//	r.log.Errorf(errMsg)
	//	return nil, fmt.Errorf(errMsg)
	//}
	//return transactionInfoBytes, nil

}

func (r *BlockRuntime) GetFullBlockByHeight(txSimContext protocol.TxSimContext, params map[string][]byte) ([]byte,
	error) {
	var errMsg string
	var err error
	var block *commonPb.Block
	var blockWithRWSetBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(params, paramNameBlockHeight); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var blockWithRWSet *storage.BlockWithRWSet
	if blockWithRWSet, err = r.getFullBlockByHeight(store, chainId, param.height); err != nil {
		return nil, err
	}

	if block, err = checkRoleAndFilterBlockTxs(blockWithRWSet.Block, txSimContext); err != nil {
		return nil, err
	}

	blockWithRWSet = &storage.BlockWithRWSet{
		Block:          block,
		TxRWSets:       blockWithRWSet.TxRWSets,
		ContractEvents: blockWithRWSet.ContractEvents,
	}

	blockWithRWSetBytes, err = blockWithRWSet.Marshal()
	if err != nil {
		errMsg = fmt.Sprintf("marshal block with rwset failed, %s", err.Error())
		r.log.Errorf(errMsg)
		return nil, fmt.Errorf(errMsg)
	}

	return blockWithRWSetBytes, nil
}

func (r *BlockRuntime) GetBlockHeightByTxId(txSimContext protocol.TxSimContext, params map[string][]byte) ([]byte,
	error) {
	var err error

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(params, paramNameTxId); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var blockHeight uint64
	blockHeight, err = r.getBlockHeightByTxId(store, chainId, param.txId)
	if err != nil {
		return nil, err
	}

	resultBlockHeight := strconv.FormatInt(int64(blockHeight), 10)
	return []byte(resultBlockHeight), nil
}

func (r *BlockRuntime) GetBlockHeightByHash(txSimContext protocol.TxSimContext, params map[string][]byte) ([]byte,
	error) {
	var err error
	var blockHeight uint64

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(params, paramNameBlockHash); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId
	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	blockHeight, err = r.getBlockHeightByHash(store, chainId, param.hash)
	if err != nil {
		return nil, err
	}

	resultBlockHeight := strconv.FormatInt(int64(blockHeight), 10)
	return []byte(resultBlockHeight), nil
}

func (r *BlockRuntime) GetBlockHeaderByHeight(txSimContext protocol.TxSimContext, params map[string][]byte) ([]byte,
	error) {
	var err error
	var errMsg string
	var blockHeaderBytes []byte

	// check params
	var param *BlockRuntimeParam
	if param, err = r.validateParams(params, paramNameBlockHeight); err != nil {
		return nil, err
	}

	chainId := txSimContext.GetTx().Payload.ChainId

	store := txSimContext.GetBlockchainStore()
	if store == nil {
		return nil, errStoreIsNil
	}

	var blockHeader *commonPb.BlockHeader
	if blockHeader, err = r.getBlockHeaderByHeight(store, chainId, param.height); err != nil {
		return nil, err
	}

	blockHeaderBytes, err = blockHeader.Marshal()
	if err != nil {
		errMsg = fmt.Sprintf("block header marshal err is %s ", err.Error())
		r.log.Error(errMsg)
		return nil, fmt.Errorf(errMsg)
	}

	return blockHeaderBytes, nil
}

func (r *BlockRuntime) getChainNodeInfo(provider protocol.ChainNodesInfoProvider, chainId string) (
	[]*discoveryPb.Node, error) {
	nodeInfos, err := provider.GetChainNodesInfo()
	if err != nil {
		r.log.Errorf("get chain node info failed, [chainId:%s], %s", chainId, err.Error())
		return nil, fmt.Errorf("get chain node info failed failed, %s", err)
	}
	nodes := make([]*discoveryPb.Node, len(nodeInfos))
	for i, nodeInfo := range nodeInfos {
		nodes[i] = &discoveryPb.Node{
			NodeId:      nodeInfo.NodeUid,
			NodeAddress: strings.Join(nodeInfo.NodeAddress, ","),
			NodeTlsCert: nodeInfo.NodeTlsCert,
		}
	}
	return nodes, nil
}

func (r *BlockRuntime) getBlockByHeight(store protocol.BlockchainStore, chainId string, height uint64) (
	*commonPb.Block, error) {
	var (
		block *commonPb.Block
		err   error
	)

	if height == math.MaxUint64 {
		block, err = store.GetLastBlock()
	} else {
		block, err = store.GetBlock(height)
	}
	err = r.handleError(block, err, chainId)
	return block, err
}

func (r *BlockRuntime) getFullBlockByHeight(store protocol.BlockchainStore, chainId string, height uint64) (
	*storage.BlockWithRWSet, error) {
	var (
		lastBlock      *commonPb.Block
		blockWithRWSet *storage.BlockWithRWSet
		err            error
	)

	if height == math.MaxUint64 {
		lastBlock, err = store.GetLastBlock()
		if err != nil {
			err = r.handleError(lastBlock, err, chainId)
			return nil, err
		}

		blockWithRWSet, err = store.GetBlockWithRWSets(lastBlock.Header.BlockHeight)
	} else {
		blockWithRWSet, err = store.GetBlockWithRWSets(height)
	}

	err = r.handleError(blockWithRWSet, err, chainId)
	return blockWithRWSet, err
}

func (r *BlockRuntime) getBlockHeaderByHeight(store protocol.BlockchainStore, chainId string, height uint64) (
	*commonPb.BlockHeader, error) {
	var (
		lastBlock   *commonPb.Block
		blockHeader *commonPb.BlockHeader
		err         error
	)

	if height == math.MaxUint64 {
		lastBlock, err = store.GetLastBlock()
		if err != nil {
			err = r.handleError(lastBlock, err, chainId)
			return nil, err
		}

		blockHeader, err = store.GetBlockHeaderByHeight(lastBlock.Header.BlockHeight)
	} else {
		blockHeader, err = store.GetBlockHeaderByHeight(height)
	}

	err = r.handleError(blockHeader, err, chainId)
	return blockHeader, err
}
func (r *BlockRuntime) getBlockHeightByTxId(store protocol.BlockchainStore, chainId string, txId string) (
	uint64, error) {
	height, err := store.GetTxHeight(txId)
	err = r.handleError(BlockHeight(height), err, chainId)
	return height, err
}

func (r *BlockRuntime) getBlockHeightByHash(store protocol.BlockchainStore, chainId string, hash string) (
	uint64, error) {
	hashBytes, err := hex.DecodeString(hash)
	if err != nil {
		r.log.Errorf("decode hash failed, [hash:%s], %s", hash, err.Error())
		return 0, fmt.Errorf("decode hash failed, %s", err)
	}
	height, err := store.GetHeightByHash(hashBytes)
	err = r.handleError(BlockHeight(height), err, chainId)
	return height, err
}

func (r *BlockRuntime) getBlockByHash(store protocol.BlockchainStore, chainId string, hash string) (
	*commonPb.Block, error) {
	hashBytes, err := hex.DecodeString(hash)
	if err != nil {
		r.log.Errorf("decode hash failed, [hash:%s], %s", hash, err.Error())
		return nil, fmt.Errorf("decode hash failed, %s", err)
	}
	block, err := store.GetBlockByHash(hashBytes)
	err = r.handleError(block, err, chainId)
	return block, err
}

func (r *BlockRuntime) getBlockByTxId(store protocol.BlockchainStore, chainId string, txId string) (
	*commonPb.Block, error) {
	block, err := store.GetBlockByTx(txId)
	err = r.handleError(block, err, chainId)
	return block, err
}

func (r *BlockRuntime) getLastConfigBlock(store protocol.BlockchainStore, chainId string) (
	*commonPb.Block, error) {
	block, err := store.GetLastConfigBlock()
	err = r.handleError(block, err, chainId)
	return block, err
}

func (r *BlockRuntime) getTxWithInfoByTxId(store protocol.BlockchainStore, chainId string, txId string) (
	*commonPb.TransactionInfo, error) {
	tx, err := store.GetTxWithInfo(txId)
	err = r.handleErrorSpecTypeName(tx, err, "transaction", chainId)
	return tx, err
}
func (r *BlockRuntime) getTxInfoWithRWSetByTxId(store protocol.BlockchainStore, chainId string, txId string) (
	*commonPb.TransactionInfoWithRWSet, error) {
	tx, err := store.GetTxInfoWithRWSet(txId)
	err = r.handleErrorSpecTypeName(tx, err, "transaction", chainId)
	return tx, err
}
func (r *BlockRuntime) getTxRWSetsByBlock(store protocol.BlockchainStore, chainId string, block *commonPb.Block) (
	[]*commonPb.TxRWSet, error) {
	var txRWSets []*commonPb.TxRWSet
	for _, tx := range block.Txs {
		txRWSet, err := store.GetTxRWSet(tx.Payload.TxId)
		if err != nil {
			r.log.Errorf("get txRWset from store failed, [chainId:%s|txId:%s], %s",
				chainId, tx.Payload.TxId, err.Error())
			return nil, fmt.Errorf("get txRWset failed, %s", err)
		}
		if txRWSet == nil { //数据库未找到记录，这不正常，记录日志，初始化空实例
			r.log.Errorf("not found rwset data in database by txid=%d, please check database", tx.Payload.TxId)
			txRWSet = &commonPb.TxRWSet{}
		}
		txRWSets = append(txRWSets, txRWSet)
	}
	return txRWSets, nil
}

func (r *BlockRuntime) GetArchiveBlockHeight(context protocol.TxSimContext, params map[string][]byte) ([]byte, error) {
	blockHeight := strconv.FormatInt(int64(context.GetBlockchainStore().GetArchivedPivot()), 10)

	r.log.Infof("get archive block height success blockHeight[%s] ", blockHeight)
	return []byte(blockHeight), nil
}

func (r *BlockRuntime) handleError(value interface{}, err error, chainId string) error {
	tn := strings.Split(fmt.Sprintf("%T", value), ".")
	typeName := strings.ToLower(tn[len(tn)-1])
	return r.handleErrorSpecTypeName(value, err, typeName, chainId)
}

func (r *BlockRuntime) handleErrorSpecTypeName(value interface{}, err error, typeName, chainId string) error {
	if err != nil {
		r.log.Errorf("get %s from store failed, [chainId:%s], %s", typeName, chainId, err.Error())
		return fmt.Errorf("get %s failed, %s", typeName, err)
	}
	vi := reflect.ValueOf(value)
	if vi.Kind() == reflect.Ptr && vi.IsNil() {
		errMsg := fmt.Errorf("no such %s, chainId:%s", typeName, chainId)
		r.log.Warnf(errMsg.Error())
		return errMsg
	}
	return nil
}

func (r *BlockRuntime) validateParams(parameters map[string][]byte, keyNames ...string) (*BlockRuntimeParam, error) {
	return r.validateParamsAndLen(parameters, true, keyNames...)
}
func (r *BlockRuntime) validateParamsAndLen(parameters map[string][]byte, checkParamLen bool, keyNames ...string) (
	*BlockRuntimeParam, error) {
	var (
		errMsg string
		err    error
	)
	if checkParamLen && len(parameters) != len(keyNames) {
		errMsg = fmt.Sprintf("invalid params len, need [%s]", strings.Join(keyNames, "|"))
		r.log.Error(errMsg)
		return nil, errors.New(errMsg)
	}
	param := &BlockRuntimeParam{}
	for _, keyName := range keyNames {
		switch keyName {
		case paramNameBlockHeight:
			value, _ := r.getValue(parameters, paramNameBlockHeight)
			if value == "-1" { //接收-1作为高度参数，用于表示最新高度，系统内部用MaxUint64表示最新高度
				param.height = math.MaxUint64
			} else {
				param.height, err = strconv.ParseUint(value, 10, 64)
			}
		case paramNameWithRWSet:
			param.withRWSet, err = r.getValue(parameters, paramNameWithRWSet)
		case paramNameBlockHash:
			param.hash, err = r.getValue(parameters, paramNameBlockHash)
		case paramNameTxId:
			param.txId, err = r.getValue(parameters, paramNameTxId)
		}
		if err != nil {
			return nil, err
		}
	}
	return param, nil
}

func (r *BlockRuntime) getValue(parameters map[string][]byte, key string) (string, error) {
	value, ok := parameters[key]
	if !ok {
		errMsg := fmt.Sprintf("miss params %s", key)
		r.log.Error(errMsg)
		return "", errors.New(errMsg)
	}
	return string(value), nil
}

func checkRoleAndFilterBlockTxs(block *commonPb.Block, txSimContext protocol.TxSimContext) (*commonPb.Block, error) {
	var (
		reqSender protocol.Role
		err       error
		ac        protocol.AccessControlProvider
	)

	ac, err = txSimContext.GetAccessControl()
	if err != nil {
		return block, err
	}

	reqSender, err = utils.GetRoleFromTx(txSimContext.GetTx(), ac)
	if err != nil {
		return block, err
	}

	reqSenderOrgId := txSimContext.GetTx().Sender.Signer.OrgId
	if reqSender == protocol.RoleLight {
		newBlock := utils.FilterBlockTxs(reqSenderOrgId, block)
		return newBlock, nil
	}

	return block, nil
}

//func checkRoleAndGenerateTransactionInfo(txSimContext protocol.TxSimContext,
//	transactionInfo *commonPb.TransactionInfo) (*commonPb.TransactionInfo, error) {
//	var (
//		reqSender protocol.Role
//		err       error
//		ac        protocol.AccessControlProvider
//	)
//	tx := transactionInfo.Transaction
//
//	if ac, err = txSimContext.GetAccessControl(); err != nil {
//		return nil, err
//	}
//
//	if reqSender, err = utils.GetRoleFromTx(txSimContext.GetTx(), ac); err != nil {
//		return nil, err
//	}
//
//	if reqSender == protocol.RoleLight {
//		if tx.Sender.Signer.OrgId != txSimContext.GetTx().Sender.Signer.OrgId {
//			newTransactionInfo := &commonPb.TransactionInfo{
//				Transaction: nil,
//				BlockHeight: transactionInfo.BlockHeight,
//				BlockHash:   transactionInfo.BlockHash,
//				TxIndex:     transactionInfo.TxIndex,
//			}
//			return newTransactionInfo, nil
//		}
//	}
//	return transactionInfo, nil
//}
